@@ -859,6 +859,20 @@ const char *bt_compidtostr(int compid)
 		return "Nimble Devices Oy";
 	case 298:
 		return "Changzhou Yongse Infotech Co., Ltd";
+	case 299:
+		return "SportIQ";
+	case 300:
+		return "TEMEC Instruments B.V.";
+	case 301:
+		return "Sony Corporation";
+	case 302:
+		return "ASSA ABLOY";
+	case 303:
+		return "Clarion Co., Ltd.";
+	case 304:
+		return "Warehouse Innovations";
+	case 305:
+		return "Cypress Semiconductor Corporation";
 	case 65535:
 		return "internal use";
 	default:
