@@ -30,20 +30,21 @@
 #include <ctype.h>
 #include <stdlib.h>
 #include <stdint.h>
+#include <stdbool.h>
 #include <glib.h>
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/hci.h>
 #include <bluetooth/sdp.h>
 
-#include "glib-helper.h"
+#include "uuid-helper.h"
 #include "eir.h"
 
 #define EIR_OOB_MIN (2 + 6)
 
 void eir_data_free(struct eir_data *eir)
 {
-	g_slist_free_full(eir->services, g_free);
+	g_slist_free_full(eir->services, free);
 	eir->services = NULL;
 	g_free(eir->name);
 	eir->name = NULL;
@@ -53,9 +54,10 @@ void eir_data_free(struct eir_data *eir)
 	eir->randomizer = NULL;
 }
 
-static void eir_parse_uuid16(struct eir_data *eir, void *data, uint8_t len)
+static void eir_parse_uuid16(struct eir_data *eir, const void *data,
+								uint8_t len)
 {
-	uint16_t *uuid16 = data;
+	const uint16_t *uuid16 = data;
 	uuid_t service;
 	char *uuid_str;
 	unsigned int i;
@@ -65,13 +67,16 @@ static void eir_parse_uuid16(struct eir_data *eir, void *data, uint8_t len)
 		service.value.uuid16 = bt_get_le16(uuid16);
 
 		uuid_str = bt_uuid2string(&service);
+		if (!uuid_str)
+			continue;
 		eir->services = g_slist_append(eir->services, uuid_str);
 	}
 }
 
-static void eir_parse_uuid32(struct eir_data *eir, void *data, uint8_t len)
+static void eir_parse_uuid32(struct eir_data *eir, const void *data,
+								uint8_t len)
 {
-	uint32_t *uuid32 = data;
+	const uint32_t *uuid32 = data;
 	uuid_t service;
 	char *uuid_str;
 	unsigned int i;
@@ -81,13 +86,16 @@ static void eir_parse_uuid32(struct eir_data *eir, void *data, uint8_t len)
 		service.value.uuid32 = bt_get_le32(uuid32);
 
 		uuid_str = bt_uuid2string(&service);
+		if (!uuid_str)
+			continue;
 		eir->services = g_slist_append(eir->services, uuid_str);
 	}
 }
 
-static void eir_parse_uuid128(struct eir_data *eir, uint8_t *data, uint8_t len)
+static void eir_parse_uuid128(struct eir_data *eir, const uint8_t *data,
+								uint8_t len)
 {
-	uint8_t *uuid_ptr = data;
+	const uint8_t *uuid_ptr = data;
 	uuid_t service;
 	char *uuid_str;
 	unsigned int i;
@@ -98,6 +106,8 @@ static void eir_parse_uuid128(struct eir_data *eir, uint8_t *data, uint8_t len)
 		for (k = 0; k < 16; k++)
 			service.value.uuid128.data[k] = uuid_ptr[16 - k - 1];
 		uuid_str = bt_uuid2string(&service);
+		if (!uuid_str)
+			continue;
 		eir->services = g_slist_append(eir->services, uuid_str);
 		uuid_ptr += 16;
 	}
@@ -126,19 +136,21 @@ static char *name2utf8(const uint8_t *name, uint8_t len)
 	return g_strdup(utf8_name);
 }
 
-int eir_parse(struct eir_data *eir, uint8_t *eir_data, uint8_t eir_len)
+void eir_parse(struct eir_data *eir, const uint8_t *eir_data, uint8_t eir_len)
 {
 	uint16_t len = 0;
 
 	eir->flags = -1;
+	eir->tx_power = 127;
 
 	/* No EIR data to parse */
 	if (eir_data == NULL)
-		return 0;
+		return;
 
 	while (len < eir_len - 1) {
 		uint8_t field_len = eir_data[0];
-		uint8_t data_len, *data = &eir_data[2];
+		const uint8_t *data;
+		uint8_t data_len;
 
 		/* Check for the end of EIR */
 		if (field_len == 0)
@@ -150,6 +162,7 @@ int eir_parse(struct eir_data *eir, uint8_t *eir_data, uint8_t eir_len)
 		if (len > eir_len)
 			break;
 
+		data = &eir_data[2];
 		data_len = field_len - 1;
 
 		switch (eir_data[1]) {
@@ -186,6 +199,12 @@ int eir_parse(struct eir_data *eir, uint8_t *eir_data, uint8_t eir_len)
 			eir->name_complete = eir_data[1] == EIR_NAME_COMPLETE;
 			break;
 
+		case EIR_TX_POWER:
+			if (data_len < 1)
+				break;
+			eir->tx_power = (int8_t) data[0];
+			break;
+
 		case EIR_CLASS_OF_DEV:
 			if (data_len < 3)
 				break;
@@ -210,12 +229,20 @@ int eir_parse(struct eir_data *eir, uint8_t *eir_data, uint8_t eir_len)
 				break;
 			eir->randomizer = g_memdup(data, 16);
 			break;
+
+		case EIR_DEVICE_ID:
+			if (data_len < 8)
+				break;
+
+			eir->did_source = data[0] | (data[1] << 8);
+			eir->did_vendor = data[2] | (data[3] << 8);
+			eir->did_product = data[4] | (data[5] << 8);
+			eir->did_version = data[6] | (data[7] << 8);
+			break;
 		}
 
 		eir_data += field_len + 1;
 	}
-
-	return 0;
 }
 
 int eir_parse_oob(struct eir_data *eir, uint8_t *eir_data, uint16_t eir_len)
@@ -236,7 +263,7 @@ int eir_parse_oob(struct eir_data *eir, uint8_t *eir_data, uint16_t eir_len)
 
 	/* optional OOB EIR data */
 	if (eir_len > 0)
-		return eir_parse(eir, eir_data, eir_len);
+		eir_parse(eir, eir_data, eir_len);
 
 	return 0;
 }
@@ -249,7 +276,7 @@ static void eir_generate_uuid128(sdp_list_t *list, uint8_t *ptr,
 	int i, k, uuid_count = 0;
 	uint16_t len = *eir_len;
 	uint8_t *uuid128;
-	gboolean truncated = FALSE;
+	bool truncated = false;
 
 	/* Store UUIDs in place, skip 2 bytes to write type and length later */
 	uuid128 = ptr + 2;
@@ -264,7 +291,7 @@ static void eir_generate_uuid128(sdp_list_t *list, uint8_t *ptr,
 
 		/* Stop if not enough space to put next UUID128 */
 		if ((len + 2 + SIZEOF_UUID128) > HCI_MAX_EIR_LENGTH) {
-			truncated = TRUE;
+			truncated = true;
 			break;
 		}
 
@@ -302,7 +329,7 @@ static void eir_generate_uuid128(sdp_list_t *list, uint8_t *ptr,
 }
 
 int eir_create_oob(const bdaddr_t *addr, const char *name, uint32_t cod,
-			uint8_t *hash, uint8_t *randomizer,
+			const uint8_t *hash, const uint8_t *randomizer,
 			uint16_t did_vendor, uint16_t did_product,
 			uint16_t did_version, uint16_t did_source,
 			sdp_list_t *uuids, uint8_t *data)
@@ -313,7 +340,7 @@ int eir_create_oob(const bdaddr_t *addr, const char *name, uint32_t cod,
 	uint16_t eir_total_len;
 	uint16_t uuid16[HCI_MAX_EIR_LENGTH / 2];
 	int i, uuid_count = 0;
-	gboolean truncated = FALSE;
+	bool truncated = false;
 	size_t name_len;
 
 	eir_total_len =  sizeof(uint16_t) + sizeof(bdaddr_t);
@@ -408,7 +435,7 @@ int eir_create_oob(const bdaddr_t *addr, const char *name, uint32_t cod,
 		/* Stop if not enough space to put next UUID16 */
 		if ((eir_optional_len + 2 + sizeof(uint16_t)) >
 				HCI_MAX_EIR_LENGTH) {
-			truncated = TRUE;
+			truncated = true;
 			break;
 		}
 
