@@ -37,13 +37,14 @@
 #include <sys/param.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
+#include <sys/stat.h>
 
 #include <bluetooth/bluetooth.h>
 #include <bluetooth/hci.h>
 #include <bluetooth/hci_lib.h>
 
-#include "textfile.h"
-#include "csr.h"
+#include "src/textfile.h"
+#include "tools/csr.h"
 
 static struct hci_dev_info di;
 static int all;
